#-----------------------------------------------------------------
# read BT1 multidetector data as a pseudo-single detector file
# process .raw files throug gformat and .bt1 files through 
# gformat and proprep -- which are not available for Windows
#------- define a command line option -----------------------------
if {$tcl_platform(platform) != "windows"} {
    # command line option 
    lappend command(cmdopt) -1D -1d
    # proc to use 
    set command(-1D) readgformatdata
    set command(-1d) readgformatdata
    #--------define a dialog box entry
    # menu label
    lappend command(readtypes) "BT-1 pseudo-1D"
    # proc for this entry
    lappend command(readproc) ReadBT1P
    # allowed data types
    lappend command(filterlist) {bt1 raw bt1gz}
    # definitions for these data types
    set command(ReadBT1P_bt1_type) "BT-1 as-collected"
    set command(ReadBT1P_bt1gz_type) "BT-1 as-collected compressed"
    set command(ReadBT1P_raw_type) "BT-1 proprep"
}

proc ReadBT1P {file} {
    global command
    if {$file == ""} return
    pleasewait "reading file $file"
    set ret [readgformatdata $file]
    donewait
    if {$ret != ""} {return $ret}
    showlastentry $command(read_filelist)
}

proc readgformatdata {filename} {
    global bt1 graph
    set lnum 0
    # signal errors by quitting
    if [catch {

	# is this a bt1 file? Check extensions.
	set ext [file extension $filename]
	if {$ext == ".bt1"} {
	    puts "opening BT1 file $filename"
	    set fp [open "| $bt1(raw_fltr) < $filename | $bt1(merge_fltr)" r] 
	} elseif {$ext == ".bt1gz"} {
	    puts "opening compressed BT1 file $filename"
	    set fp [open "| gunzip < $filename | $bt1(raw_fltr) | $bt1(merge_fltr)" r] 
	} elseif {$ext == ".raw"} {
	    puts "opening raw file $filename"
	    set fp [open "| $bt1(merge_fltr) < $filename" r] 
	} else {
	    return "file $filename has an invalid extension for BT1 data"
	}
	
	# output from gformat does not need to be checked as closely as 
	# the general GSAS case it will always be CONST - ESD format

	# 1st line is title
	incr lnum
	set eof [gets $fp title]
	if {$eof < 0} {return "end-of-file before BANK found"}
	# second line may be "Instrument parameter" line
	incr lnum
	set eof [gets $fp line]
	if {$eof < 0} {return "end-of-file before BANK found"} 
	# skip line if it does not contain BANK starting in the 1st column
	if {[string range $line 0 3] != "BANK"} {
	    incr lnum
	    set eof [gets $fp line]
	    if {$eof < 0} {return "end-of-file before BANK found"} 
	}
	set start [lindex $line 5]
	set step [lindex $line 6]
	set npts [lindex $line 2]
	if {[lindex $line 9] == ""} {set esdflag 0} {set esdflag 1}
	set ptnum 0
	set data [file root [file tail $filename]] 
	initdata $data
	global ${data}
	set ${data}(title) [string trim [string range $title 0 45]]
	set ${data}(xlabel) "2theta"
	set ${data}(xunits) "2theta"
	set ${data}(ylabel) "scaled counts"

	set ptsread 0
	# are there missing points?
	set ${data}(skip) 0
	if {$esdflag} {
	    while {$ptnum < $npts} {
		incr lnum
		set eof [gets $fp line]
		if {$eof < 0} {return "end-of-file before all data read"} 
		if {$lnum > 20000} {
		    return "Read too many lines. Something is wrong!"
		}
		foreach num {0 16 32 48 64} {
		    set num7 [expr $num + 7]
		    set count [string trim [string range $line $num $num7]]
		    set num8 [expr $num + 8]
		    set num15 [expr $num + 15]
		    set esd [string trim [string range $line $num8 $num15]]
		    if {$count >= 0 && $esd > 0} {
			lappend ${data}(x) \
				[expr ($start+$step*$ptnum)/100.]
			lappend ${data}(y) $count
			lappend ${data}(esd) $esd
			incr ptsread
		    } else {
			# missing points
			set ${data}(skip) 1
		    }
		    incr ptnum
		}
	    }
	} else {
	    return "Incorrect format -- no ESD flag!"
	}
	
    } errmsg] {
	# try to process the data we have read
	catch {resetdata $data}
	return "BT1 error. Error reading line $lnum: $errmsg"
    }    
    resetdata $data
    lappend graph(plotlist) $data
    catch {close $fp} test
    puts "$ptsread valid points read"
    return {}
}
